import numpy as np
import scipy, os, glob
from scipy.signal import butter,lfilter
from scipy.ndimage.filters import gaussian_filter1d
import matplotlib.pyplot as plt
from matplotlib.pyplot import mlab
import xml.etree.ElementTree
samplingRate=30000.
#=================================================================================================
#------------operations on continuous traces-------------------------------------
#=================================================================================================

npix_p3_reference_channels = np.array([ 36,  75, 112, 151, 188, 227, 264, 303, 340, 379])
npix_p2_reference_channels = np.array([1,18,33,50,65,82,97,114,99])
skip_channels = npix_p3_reference_channels #default to phase 3 reference channels
def get_chunk(mm,start,end,channels,sampling_rate=30000,remove_offset=False):
	chunk = mm[int(start*sampling_rate*int(channels)):int(np.floor(end*sampling_rate*(int(channels))))]
	chunk = np.reshape(chunk,(int(channels),-1),order='F')  * 0.195
	# print(np.shape(chunk))
	if remove_offset:
		for i in np.arange(channels):
			chunk[i,:] = chunk[i,:] - np.mean(chunk[i,:])
	return chunk

#filter a bit of continuous data. uses butterworth filter.
def filterTrace(trace, low, high, sampleHz, order):
    low = float(low)
    high = float(high)
    nyq = 0.5 * sampleHz
    low = low / nyq
    high = high / nyq
    b, a = butter(order, [low, high], btype='band')
    filtered = lfilter(b, a, trace)
    return filtered
    
#developmental filter version. not used.
def filterTrace_hard(trace, low, high, sampleHz, order):
    low = float(low)
    high = float(high)
    nyq = 0.5 * sampleHz
    low = low / nyq
    high = high / nyq
    scipy.signal.band_stop_obj()
    b, a = butter(order, [low, high], btype='band')
    filtered = lfilter(b, a, trace)
    scipy.signal.lfilter()
    return filtered

#wrapper for filtering continous data of different forms.
#data can be a single continuous trace, a dictionary containing a key called 'data' whose value is a continous trace, or a dictionary of traces, or a dicit
def filtr(data,low, high, sampleHz, order):
    if type(data) is dict:
        if 'data' in data.keys():
            return filterTrace(data['data'],low, high, sampleHz, order)
        else:
            out = {}
            for i,key in enumerate(data.keys()):
                out[key] = data[key]
                out[key]['data']= filterTrace(data[key]['data'],low, high, sampleHz, order)
            return out
    else:
        return filterTrace(data,low, high, sampleHz, order)

#notch filter a continous trace by filtering in a narrow range and subtracting that from the input trace.    
def notch(data,freq, sampleHz):
    order = 1
    low = freq-2
    high = freq +2
    if type(data) is dict:
        if 'data' in data.keys():
            return data['data'] - filterTrace(data['data'],low, high, sampleHz, order)
        else:
            out = {}
            for i,key in enumerate(data.keys()):
                out[key] = data[key]
                out[key]['data']= data[key]['data'] - filterTrace(data[key]['data'],low, high, sampleHz, order)
            return out
    else:
        return data - filterTrace(data,low, high, sampleHz, order)

#average a continuous trace around a set of timestamps 
def average_trials(data,timestamps,window,sampleFreq=25000.):
    alltrials = np.zeros((len(timestamps),window*sampleFreq))
    average = np.zeros(window*sampleFreq)
    skipped = 0
    for i,onset in enumerate(timestamps):
        average += data[onset:onset+window*sampleFreq]#-np.mean(data[onset:onset-500])
        alltrials[i,:] = data[onset:onset+window*sampleFreq]#-np.mean(data[onset:onset-500])
        
#        if np.max(np.abs(data[onset:onset+window*sampleFreq]-np.mean(data[onset:onset+5000]))) < 40000000.0:
#            average += data[onset:onset+window*sampleFreq]-np.mean(data[onset:onset+5000])
#            alltrials[i,:] = data[onset:onset+window*sampleFreq]-np.mean(data[onset:onset+5000])
#        else:
#            skipped += 1
#            print 'skipped trial: '+str(i+1)
#            alltrials[i,:] = data[onset:onset+window*sampleFreq]-np.mean(data[onset:onset+5000])
    return alltrials,average/float(len(timestamps-skipped))   

#average all continuous traces in an array around a set of timestamps 
def average_trials_array(data,timestamps,window,output='avg'):
    avgs = {}
    alltrials={}
    for i,key in enumerate(data.keys()):
    	if 'data' in data[key].keys():
	        avgs[key]={}
	        alltrials[key]={}
	        alltrials[key]['data'],avgs[key]['data'] = average_trials(data[key]['data'],timestamps,window)
    if output == 'trials':
        return alltrials
    if output == 'both':
        return (alltrials,avgs)
    if output=='avg':
        return avgs
	

#note: this CSD code does not work! -dan
def CSD_1D(data,channelmap=[],prefix='100_CH',point=1000):
    if channelmap == []:
        channelmap = data.keys()
    elec_pos = []
    pots=[]    
    for i,key in enumerate(channelmap[0]):
        key = prefix+str(key).replace(prefix,'')
        pots.append([data[key]['data'][point]])
        elec_pos.append([(i+i)/2])
    pots=np.array(pots)
    elec_pos=np.array(elec_pos)
    params = {
        'xmin': 0,
        'xmax': 65.0,
        'source_type': 'step',
        'n_sources': 64,
        'sigma': 0.1
        }
    k = KCSD(elec_pos, pots, params)
    k.estimate_pots()
    k.estimate_csd()
    k.plot_all()

#note: this CSD code does not work! -dan    
def CSD_1D_time(data,channelmap=[],prefix='100_CH',point=1000):
    if channelmap == []:
        channelmap = data.keys()
    numPoints = len(data[data.keys()[0]]['data'])
    out_csd = np.zeros((len(data.keys()),numPoints))
    out_pots = np.zeros((len(data.keys()),numPoints))
    for point in range(numPoints):
        print(point)
        elec_pos = []
        pots=[]    
        for i,key in enumerate(channelmap[0]):
            key = prefix+str(key).replace(prefix,'')
            pots.append([data[key]['data'][point]])
            elec_pos.append([i+i])
        pots=np.array(pots)
        elec_pos=np.array(elec_pos)
        params = {
            'xmin': 0,
            'xmax': 130.0,
            'source_type': 'step',
            'n_sources': 128,
            'sigma': 0.2,
            }
        k = KCSD(elec_pos, pots, params)
        k.estimate_pots()
        k.estimate_csd()
        out_csd[0:np.shape(k.solver.estimated_csd)[0],point]= k.solver.estimated_csd[:,0]
        out_pots[0:np.shape(k.solver.estimated_pots)[0],point]= k.solver.estimated_pots[:,0]
    return out_csd,out_pots
        #k.plot_all()

def etree_to_dict(t):
    d = {t.tag : list(map(etree_to_dict, t.getchildren()))}
    d.update(('@' + k, v) for k, v in t.attrib.items())
    d['text'] = t.text
    return d

def get_channel_count(path,from_channel_map = True,from_templates=False):
	d = etree_to_dict(xml.etree.ElementTree.parse(os.path.join(path,'settings.xml')).getroot())
	chs =0
	if from_templates:
		return np.load(open(os.path.join(path,'templates.npy'))).shape[-1]
	if d['SETTINGS'][1]['SIGNALCHAIN'][0]['@name'] == 'Sources/Neuropix':
		for info in d['SETTINGS'][1]['SIGNALCHAIN'][0]['PROCESSOR'][:385]:
			if 'CHANNEL' in info.keys():
				if info['CHANNEL'][0]['@record'] == '1':
					chs +=1
		return chs
	if d['SETTINGS'][1]['SIGNALCHAIN'][0]['@name'] == 'Sources/Rhythm FPGA':
		if from_channel_map:
			for nm in d['SETTINGS'][1]['SIGNALCHAIN']:
				name = nm['@name']
				if name == 'Filters/Channel Map':
					#chs = np.shape(d['SETTINGS'][1]['SIGNALCHAIN'][0]['PROCESSOR'][0]['CHANNEL_INFO'])[0]
					for info in nm['PROCESSOR']:
						if 'CHANNEL' in info.keys():
							if info['CHANNEL'][0]['@record'] == '1':
								chs +=1
		else:
			for info in d['SETTINGS'][1]['SIGNALCHAIN'][0]['PROCESSOR'][:385]:
				if 'CHANNEL' in info.keys():
					if info['CHANNEL'][0]['@record'] == '1':
						chs +=1
		return chs

#returns the root mean squared of the input data    
def RMS(data,start=0,window=0,despike=False):
	start = start * samplingRate# sampling rate
	if window == 0:
		window = len(data)
	else:
		window = window * samplingRate # sampling rate
	#chunk = filterTrace(data[start:start+window], 70, 6000, 25000, 3)[200:window]
	chunk = data[int(start):int(start)+int(window)] - np.mean(data[int(start):int(start)+int(window)])
	if despike:
		chunk = despike_trace(chunk,threshold=180)
	return np.sqrt(sum(chunk**2)/float(len(chunk)))

def despike_trace(trace,threshold_sd = 2.5,**kwargs):
	if 'threshold' in kwargs.keys():
		threshold = kwargs['threshold']
	else:
		threshold = np.mean(trace)+threshold_sd*np.std(trace)
		
	spike_times_a = mlab.cross_from_below(trace,threshold)
	spike_times_b = mlab.cross_from_below(trace,-1*threshold)
	for spike_time in np.concatenate((spike_times_b,spike_times_a)):
		if spike_time > 30 and spike_time < len(trace)-30:
			trace[spike_time - 20:spike_time + 20] = 0#np.random.uniform(-1*threshold,threshold,60)
	return trace

def spikeamplitudes_trace(trace,threshold_sd = 3.0,percentile = 0.9,**kwargs):
	if 'threshold' in kwargs.keys():
		threshold = kwargs['threshold']
	else:
		threshold = np.mean(trace)+threshold_sd*np.std(trace)
		
	spike_times_a = mlab.cross_from_below(trace,threshold)
	amps=[]
	for spike_time in spike_times_a:
		if spike_time > 30 and spike_time < len(trace)-30:
			amps.extend([np.max(np.abs(trace[spike_time-30:spike_time+30]))])
	if not len(amps) > 10:
		amps= [0]
	return np.sort(amps)[int(len(amps)*percentile)]# / 5.0

#returns the peak to peak range of the input data     
def p2p(data,start=0,window=0):
    start = start * samplingRate# sampling rate
    if window == 0:
        window = len(data)
    else:
        window = window * samplingRate # sampling rate
    chunk = data[start:start+window]
    return np.max(chunk)-np.min(chunk)
    
#computes a power spectrum of the input data
#optionally, plot the computed spectrum
def b(data,start=0,window=0,plot=False,ymin=1e-24,ymax=1e8,title='',samplingRate=2500):
    start = start * samplingRate# sampling rate
    if window == 0:
        window = len(data)
    else:
        window = window * samplingRate # sampling rate
    chunk = data[start:start+window]/1e6
    ps = np.abs(np.fft.fft(chunk))**2
    time_step = 1. / samplingRate
    freqs = np.fft.fftfreq(chunk.size, time_step)
    idx = np.argsort(freqs)
    ps = scipy.signal.savgol_filter(ps,5,3)
    if plot:
        plt.plot(freqs[idx], ps[idx]);
        plt.xlim(xmin=0.01);
        plt.ylim(ymin=ymin,ymax=ymax)
        plt.xscale('log')
        plt.yscale('log')
        plt.ylabel(r'$power\/density\/\frac{V^2}{Hz}$',color='k',fontsize=18)
        plt.xlabel(r'$frequency,\/ Hz$',color='k',fontsize=24)
        plt.tick_params(axis='both', which='major', labelsize=24)#;plt.locator_params(axis='y',nbins=6)
        plt.title(title)
    return (freqs[idx], ps[idx])

def periodogram(data,start=0,window=0,plot=False,ymin=1e-24,ymax=1e8,title='',samplingRate=2500):
    start = start * samplingRate# sampling rate
    if window == 0:
        window = len(data)
    else:
        window = window * samplingRate # sampling rate
    chunk = data[start:start+window]
    f,pXX = scipy.signal.periodogram(chunk,samplingRate,nfft=samplingRate)
    pXX = scipy.signal.savgol_filter(pXX,3,1)
    if plot:
        plt.plot(f, pXX);
        plt.xlim(xmin=0.5);
        plt.ylim(ymin=ymin,ymax=ymax)
        plt.xscale('log')
        plt.yscale('log')
        plt.ylabel(r'$power\/density\/\frac{V^2}{Hz}$',color='k',fontsize=18)
        plt.xlabel(r'$frequency,\/ Hz$',color='k',fontsize=24)
        plt.tick_params(axis='both', which='major', labelsize=24)#;plt.locator_params(axis='y',nbins=6)
        plt.title(title)
    return (f, pXX)

def welch_power(data,samplingRate=2500,start=0,window=0,plot=False,ymin=1e-24,ymax=1e8,title=''):
    start = start * samplingRate# sampling rate
    if window == 0:
        window = len(data);print(window)
    else:
        window = window * samplingRate # sampling rate
    chunk = data[start:start+window]
    f,pXX = scipy.signal.welch(chunk,samplingRate,nfft=samplingRate/2)
    #pXX = scipy.signal.savgol_filter(pXX,3,1)
    if plot:
        plt.plot(f, pXX);
        plt.xlim(xmin=0.01);
        plt.ylim(ymin=ymin,ymax=ymax)
        plt.xscale('log')
        plt.yscale('log')
        plt.ylabel(r'$power\/density\/\frac{V^2}{Hz}$',color='k',fontsize=18)
        plt.xlabel(r'$frequency,\/ Hz$',color='k',fontsize=24)
        plt.tick_params(axis='both', which='major', labelsize=24)#;plt.locator_params(axis='y',nbins=6)
        plt.title(title)
    return (f, pXX)

#measure the cross-spectral coherence between two traces. 
def coherence(x,y,samplingRate = 30000,returnval=None):
    spectrum, frequencies = mlab.cohere(x,y,Fs=float(samplingRate),NFFT=int(samplingRate)/5)
    if returnval:
        if type(returnval) is float:
            return np.interp(returnval,frequencies,spectrum)
        if type(returnval) is tuple:
            return np.trapz(spectrum[np.where(frequencies==returnval[0])[0]:np.where(frequencies==returnval[1])[0]],dx=5.0)  
    else:
        return (spectrum, frequencies)

def get_surface_channel_spikeband(path,start=2.,end=10.,sampling_rate=30000,plot=False,filter_size=2,sigma=1.,filter=False,probemap=None):
	mm = np.memmap(path, dtype=np.int16, mode='r')
	print(os.path.dirname(path))
	num_channels = get_channel_count(os.path.dirname(path),from_channel_map=False)
	print(num_channels)
	chunk = get_chunk(mm,start,end,num_channels,sampling_rate)
		
	if probemap is not None:
		chunk = chunk[probemap,:]
		plt.imshow(chunk[:,:30000]);plt.gca().set_aspect(100)
		plt.figure()
		
	rms = []
	good_channels = []
	for ch in range(np.shape(chunk)[0]):
		if ch not in skip_channels:
			if filter:
				data = filtr(chunk[ch,:],300,6000,sampling_rate,3)
			else:
				data = chunk[ch,:]
			rms.extend([RMS(data)])
			good_channels.extend([ch])
			
	threshold = np.mean(gaussian_filter1d(rms,filter_size)[::-1][:5])+np.std(gaussian_filter1d(rms,filter_size)[::-1][:5])*sigma		#assumes the last 5 are out of the brain; uses the mean + sd of these 5 as the threshold for pial surface
	
	# print(np.where(np.array(rms)<8.))
	# print(good_channels[np.where(np.array(rms)<8.)[0].astype(int)])
	if plot:
		plt.plot(good_channels,gaussian_filter1d(rms,filter_size))
		plt.gca().axhline(threshold,color='r')
		plt.xlabel('channel number')
		plt.ylabel('spike band RMS')
		#print(np.where(np.array(rms)<6.))
	del mm
	try:
		surface_channel = good_channels[mlab.cross_from_above(gaussian_filter1d(rms,filter_size),threshold)[0]]
		return surface_channel
	except:
		return None

def get_surface_channel_gamma(path,start=2.,end=10.,sampling_rate=2500,plot=False):
	mm = np.memmap(path, dtype=np.int16, mode='r')
	num_channels = get_channel_count(os.path.dirname(path))
	chunk = get_chunk(mm,start,end,num_channels,sampling_rate)
	
	gm = []
	good_channels = []
	for ch in range(np.shape(chunk)[0]):
		if ch not in skip_channels:
			f,pXX = welch_power(chunk[ch,:],start=2,window=8)
			gm.extend([pXX[np.where(f>40.)[0][0]]])
			good_channels.extend([ch])
	threshold = np.max(gm[::-1][:5])	#assumes the last 5 are out of the brain; uses the max gamma on these channels as the threshold
	surface_channel = good_channels[mlab.cross_from_above(gaussian_filter1d(gm,0),threshold)[0]]

	if plot:
		plt.plot(good_channels,gaussian_filter1d(gm,2))
		plt.gca().axhline(threshold,color='r')
	del mm
	return surface_channel

def get_surface_channel_freq(path,frequency_range=[1,5],start=2.,end=10.,sampling_rate=2500,filter_size=2,sigma=2.,plot=False,filter=False,probemap=None):
	mm = np.memmap(path, dtype=np.int16, mode='r')
	num_channels =  get_channel_count(os.path.dirname(path),from_channel_map=False)
	chunk = get_chunk(mm,start,end,num_channels,sampling_rate)
	if probemap is not None:
		chunk = chunk[probemap,:]
	gm = []
	good_channels = []
	for ch in range(np.shape(chunk)[0]):
		if ch not in skip_channels:
			if filter:
				data = filtr(chunk[ch,:],0.1,300,sampling_rate,3)
			else:
				data = chunk[ch,:]
			f,pXX = welch_power(chunk[ch,:],start=2,window=8)
			gm.extend([np.mean(pXX[np.where((f>frequency_range[0])&(f<frequency_range[1]))[0]])])
			good_channels.extend([ch])
	#threshold = np.mean(gm[::-1][:5])	#assumes the last 5 are out of the brain; uses the max gamma on these channels as the threshold
	threshold = np.mean(gaussian_filter1d(gm,filter_size)[::-1][:5])+np.std(gaussian_filter1d(gm,filter_size)[::-1][:5])*sigma

	if plot:
		plt.plot(good_channels,gaussian_filter1d(gm,filter_size))
		plt.gca().axhline(threshold,color='r')
		plt.xlabel('channel number')
		plt.ylabel('power in '+str(frequency_range[0])+' to '+str(frequency_range[1])+' band')
	try:
		surface_channel = good_channels[mlab.cross_from_above(gaussian_filter1d(gm,filter_size),threshold)[-1]]
		return surface_channel
	except:
		return None
	del mm
	return surface_channel

def get_probe_freq(path,frequency_range=[1,5],start=2.,end=10.,sampling_rate=2500.,filter=False,probemap=None):
	mm = np.memmap(path, dtype=np.int16, mode='r')
	num_channels = get_channel_count(os.path.dirname(path),from_channel_map=False)
	chunk = get_chunk(mm,start,end,num_channels,sampling_rate)
	if probemap is not None:
		chunk = chunk[probemap,:]
	gm = []
	good_channels = []
	for ch in range(np.shape(chunk)[0]):
		if ch not in skip_channels:
			if filter != False:
				data = filtr(chunk[ch,:],filter[0],filter[1],sampling_rate,3)
			else:
				data = chunk[ch,:]
			f,pXX = welch_power(chunk[ch,:],start=2,window=8)
			gm.extend([np.mean(pXX[np.where((f>frequency_range[0])&(f<frequency_range[1]))[0]])])
			good_channels.extend([ch])

	del mm
	return gm

def get_probe_spikeband(path,start=2.,end=10.,sampling_rate=30000,plot=False,filter_size=2,sigma=1.,filter=False,probemap=None):
	mm = np.memmap(path, dtype=np.int16, mode='r')
	num_channels = get_channel_count(os.path.dirname(path),from_channel_map=False)
	#print num_channels
	chunk = get_chunk(mm,start,end,num_channels,sampling_rate)
		
	if probemap is not None:
		chunk = chunk[probemap,:]
		plt.imshow(chunk[:,:30000]);plt.gca().set_aspect(100)
		plt.figure()
		
	rms = []
	good_channels = []
	for ch in range(np.shape(chunk)[0]):
		if ch not in skip_channels:
			if filter:
				data = filtr(chunk[ch,:],300,6000,sampling_rate,3)
			else:
				data = chunk[ch,:]
			rms.extend([RMS(data)])
			good_channels.extend([ch])
	del mm
	return rms
#=================================================================================================

#=================================================================================================
#=======plotting=============================================================================
#=================================================================================================
def get_chunk(mm,start,end,channels,sampling_rate=30000):
	chunk = mm[int(start*sampling_rate*int(channels)):int(np.floor(end*sampling_rate*(int(channels))))]
	#print np.shape(chunk)
	return np.reshape(chunk,(int(channels),-1),order='F')  * 0.195

def get_duration(mm,channels,sampling_rate=30000.):
    chunk=np.reshape(mm,(int(channels),-1),order='F')
    return chunk.shape[1]/float(sampling_rate)

def get_spike(spks_path,times,number_channels,pre=0.015,post=0.025,sampling_rate=30000):
    mm = np.memmap(spks_path, dtype=np.int16, mode='r')
    average = np.zeros((number_channels,int(sampling_rate*(pre+post))))
    count=0
    for time in times:
        try:
            temp = traces.get_chunk(mm,time-pre,time+post,number_channels,sampling_rate=sampling_rate)
            average += temp
            count+=1
        except:
            pass
    temp = temp/float(count)
    spikes_average = np.array(average.T - np.mean(average,1).T).T
    del(mm)

def get_lfp(lfp_path,times,number_channels,pre=0.015,post=0.025,sampling_rate=2500):
    mm = np.memmap(lfp_path, dtype=np.int16, mode='r')
    average = np.zeros((number_channels,int(sampling_rate*(pre+post))))
    count=0
    for time in sub_times:
        try:
            temp = traces.get_chunk(mm,time-pre,time+post,number_channels,sampling_rate=sampling_rate)
            average += temp
            count+=1
        except:
            pass
    temp = temp/float(count)
    lfp_average = np.array(average.T - np.mean(average,1).T).T
    del(mm)

def make_range_slider(data,start,window,num_channels=384,channels = [10],sampling_rate=2500,y_spacing=500,CAR=False):
    fig,ax=plt.subplots(figsize=(20,5))
    chunk = get_chunk(data,start,start+window,num_channels,sampling_rate=2500)
    if CAR:
        chunk_CAR = np.mean(chunk, axis=0)
    x = np.linspace(int(start),(start+window),int(window*sampling_rate))
    for i,ch in enumerate(channels):
        chunkch = chunk[ch,:]
        offset = np.mean(chunkch)
        if CAR: chunkch = chunkch - chunk_CAR
        ax.plot(x,chunkch-offset+i*y_spacing,'k',lw=.5)
#=================================================================================================

def recreate_probe_timestamps_from_TTL(directory):
    probe = directory.split('-AP')[0][-1]
    recording_base = os.path.dirname(os.path.dirname(directory))

    with open(os.path.join(recording_base,'sync_messages.txt')) as f:
        lines = f.readlines()
        for line in lines:
            if 'Probe'+probe+'-AP' in line:
                cont_start_sample = int(line.split(':')[1][1:].split('\n')[0])
    f.close()

    TTL_samples = np.load(os.path.join(glob.glob(os.path.join(recording_base,'events')+'/*Probe'+probe+'*AP*')[0],'TTL','sample_numbers.npy'))[::2]
    TTL_timestamps = np.load(os.path.join(glob.glob(os.path.join(recording_base,'events')+'/*Probe'+probe+'*AP*')[0],'TTL','timestamps.npy'))[::2]

    cont_raw = np.memmap(os.path.join(directory,'continuous.dat'),dtype=np.int16)
    cont_samples = np.arange(cont_start_sample, cont_start_sample+(int(cont_raw.shape[0]/384)))
    cont_timestamps = np.zeros(int(cont_raw.shape[0]/384))

    for i,sample in enumerate(TTL_samples):
        ind = sample-cont_start_sample
        cont_samples[ind] = sample
        cont_timestamps[ind] = TTL_timestamps[i]
        if i==0:
            cont_timestamps[:ind]=np.linspace(TTL_timestamps[i]-(1/30000. * len(cont_timestamps[:ind-1]))+1/30000.,TTL_timestamps[i],len(cont_timestamps[:ind]))
            prev_ind =ind
        else:
            cont_timestamps[prev_ind:ind] = np.linspace(cont_timestamps[prev_ind]+1/30000.,TTL_timestamps[i],len(cont_timestamps[prev_ind:ind]))
            prev_ind =ind
    cont_timestamps[ind:]=np.linspace(TTL_timestamps[i]+1/30000.,TTL_timestamps[i]+len(cont_timestamps[ind:])*1/30000.,len(cont_timestamps[ind:]))

    if not os.path.exists(os.path.join(directory,'new_timestamps')):
        os.mkdir(os.path.join(directory,'new_timestamps'))
    np.save(open(os.path.join(directory,'new_timestamps','sample_numbers.npy'),'wb'),cont_samples.astype(np.int64))
    np.save(open(os.path.join(directory,'new_timestamps','timestamps.npy'),'wb'),cont_timestamps.astype(np.float64))

### jordan raw data functions

import seaborn as sns
from open_ephys.analysis import Session

def OE(path):
    'lazy wrappy to return recording session from open ephys'
    session = Session(path)
    recording = session.recordnodes[0].recordings[0]
    return recording

def load_datastream(path, probe, band='ap'):
    '''
    purpose of this is to return a data_stream object. can then access metadata and load raw data from here.
    path: path to recording
    probe: probe name (probeA, probeB, probeC)
    '''
    # normalize case for flexibility
    probe = probe.lower()
    band = band.lower()

    # Get the recording
    recording = OE(path)

    # Search for the correct data stream by examining metadata
    for data_stream in recording.continuous:
        stream_name = data_stream.metadata['stream_name'].lower().replace('-', '_')
        if stream_name == f"{probe}_{band}":
            print(f'confirming stream name: {data_stream.metadata["stream_name"]}')
            return data_stream

    print('No matching bands or probes')
    return None

def get_chunk_OE(path,
              probe, 
            stim_times,
            band = 'ap',
            pre = 100, # time in ms
            post = 500, # time in ms
            chs = np.arange(0,200,1), # channels
            median_subtraction = False,
            ):
    """
    for open ephys data
    Takes in a continuous datastream object (from open ephys) and a list of stimulation times and returns a chunk of the data

    return: data: np.array, shape = (trials, samples, channels)
    """
    
    data_stream = load_datastream(path, probe, band = band)
    sample_rate = data_stream.metadata['sample_rate']
    
    
    pre_samps = int((pre/1000 * sample_rate))
    post_samps = int((post/1000 * sample_rate))
    total_samps = pre_samps + post_samps

    n_chs = len(chs)
    
    response = np.zeros((np.shape(stim_times)[0],total_samps, len(chs)))
    stim_indices = np.searchsorted(data_stream.timestamps, stim_times)
    for i, stim in enumerate(stim_indices):
        start_index = int(stim - ((pre/1000)*sample_rate))
        end_index = int(stim + ((post/1000)*sample_rate))   
        chunk = data_stream.get_samples(start_sample_index = start_index, end_sample_index = end_index, 
                            selected_channels = chs)
        
        if median_subtraction == True:
            corrected_chunk = chunk - np.median(chunk, axis = 0) #subtract offset 
            corrected_chunk = np.subtract(corrected_chunk.T, np.median(corrected_chunk[:,chs[-10]:chs[-1]], axis=1)) #median subtraction

            response[i,:,:] = corrected_chunk.T
        else:
            response[i,:,:] = chunk - np.median(chunk, axis = 0)

    return response

def subtract_offset(data, subtraction_window = None, pre = None, post = None):
    '''
    takes a chunk of data and subtracts the median of the data from each channel

    data: np.array, shape = (trials, samples, channels)
    subtraction_window: 'pre', 'all'. If None, will use full.
    pre: pre window in ms
    post: post window in ms 
    '''
    if subtraction_window == 'pre':
        pre_samps = int((pre/1000 * 30000))
        window = (0, pre_samps)
        pre_data = data[:,window[0]:window[1],:]
        corrected_chunk = data - np.median(pre_data, axis = 1) # the samples axis
    else: # subtract the median of the entire chunk
        corrected_chunk = data - np.median(data, axis = 1)
    
    return corrected_chunk


def median_subtraction(data, channels):
    """median subtraction (not an offset). Like CAR.
    TODO: add ability to select which channels to subtract from

    Args:
        data (np.array): The data to subtract from, shape = (trials, samples, channels).
        channels (list): The channels to subtract from.

    Returns:
        np.array: The data with the median subtracted.
    """

    corrected_data = data.copy()
    corrected_data[:, :, channels] = np.subtract(corrected_data[:, :, channels].T, np.median(corrected_data[:, :, channels], axis=1)).T

    return corrected_data


def find_artifact_start(data_ch, pre, threshold):
    """Find the start of artifact based on a threshold.
    
    Args:
        data_ch: np.array, shape = ch x sample
        threshold: The threshold used to detect the artifact. Function looks for 
                   the first value that exceeds this threshold in absolute terms.
                   
    Returns:
        The index of the first data point that exceeds the threshold.
    """
    # Use the absolute value to find the first occurrence beyond the threshold, 
    # regardless of the sign of the artifact.

    return next((i for i, val in enumerate(np.abs(data_ch)) if val > threshold), int((pre+2)/1000 * 30000))

def align_data(data, pre, post, channels, threshold = 400, median_subtraction = False):
    """ Align the data to the artifact onset based on artifact start times. Optionally perform median subtraction.

    Args:
        data (np.array): The data to align, shape = (trials, samples, channels)
        pre (float): pre-stim time in ms
        post (float): post-stim time in ms
        channels (_type_): number of channels to plot
        median_subtraction (bool): whether to perform median subtraction on the data

    Returns:
        np.array: aligned_data, shape = (trials, total_samps, channels)
    """   
    starts = [find_artifact_start(data[trial, :, 0], pre = pre, threshold = threshold) for trial in range(data.shape[0])] # the sample number of when the artifact first starts 
    # the sample number of when the artifact first starts determined by the find_artifact_start function
    
    pre_samps = (int((pre/1000) * 30000)) 
    post_samps = (int((post/1000) * 30000))
    total_samps = pre_samps + post_samps
    aligned_data = np.zeros((data.shape[0], total_samps, data.shape[2]))
    
    for trial in range(data.shape[0]):
        start = starts[trial]
        chunk = data[trial, start - pre_samps:start + post_samps, :]
        if median_subtraction:
            corrected_chunk = chunk.T - np.median(chunk[:, channels-11:channels-1], axis=1)
            aligned_data[trial, :, :] = corrected_chunk.T
        else:
            aligned_data[trial, :, :] = chunk
    
    return aligned_data




def raw_heatmap(data, pre=1, post=2, dists=None, vmin=None, vmax=None, 
                save=False, save_path=None, save_type='png', title=None, ax=None):
    """
    Plots a heatmap of data with options for customization and saving.

    Args:
        data (np.array): The data to plot, shape = (trials, samples, channels).
        pre (int, optional): pre_window in ms. Defaults to 1.
        post (int, optional): post_window in ms. Defaults to 2.
        dists (np.array, optional): Distance from stimulation array, same shape as channels.
        vmin (float, optional): Minimum value for heatmap scaling. Defaults to None.
        vmax (float, optional): Maximum value for heatmap scaling. Defaults to None.
        save (bool, optional): If True, save the figure. Defaults to False.
        save_path (str, optional): Path to save the figure. Defaults to None.
        save_type (str, optional): Format of saved figure. Defaults to 'png'.
        title (str, optional): Figure title. Defaults to None.
        ax (matplotlib.axes.Axes, optional): Axes object to plot on. Defaults to None (creates new figure).

    Returns:
        matplotlib.axes.Axes: The axes object containing the plot.
    """

    # Check if an axes object was provided, if not create a new figure and axes
    if ax is None:
        fig, ax = plt.subplots(figsize=(12, 12))
        created_fig = True
    else:
        created_fig = False  # Flag to avoid creating a new figure

    channels = data.shape[2]
    data_to_plot = np.mean(data, axis=0).T  # average across trials

    time_ms = np.linspace(-pre, post, data.shape[1])

    cax = ax.imshow(data_to_plot, aspect='auto', extent=[time_ms[0], time_ms[-1], 0, channels],
                    vmax=vmax, vmin=vmin, origin='lower', cmap='viridis')

    if created_fig:
        fig.colorbar(cax, ax=ax, pad=0.20)
    ax.set_ylabel('Channel')
    ax.set_yticks(np.arange(0, channels, 10))

    if dists is not None:
        ax_dist = ax.twinx()
        ax_dist.set_ylim(ax.get_ylim())
        ax_dist.set_yticks(np.arange(0, len(dists), 25))
        ax_dist.set_yticklabels([int(d) for d in dists[::25]])
        ax_dist.set_ylabel('Distance from Stimulation (units)')
        zero_dist_channel = np.argmin(np.abs(dists))
        ax.axhline(y=zero_dist_channel, color='red', linestyle='--')

    ax.set_xlabel('Time (ms)')
    if title:
        ax.set_title(title)

    if created_fig:
        plt.tight_layout()

    if save and created_fig:
        plt.savefig(save_path, format=save_type)

    return ax



def plot_ap(path, probe, stim_times, 
                pre = 5, post = 10, 
                first_ch = 125, last_ch = 175, 
                median_subtraction = False,
            
                spike_overlay = False,
                units = None,
                title = '', 

                n_trials = 10, spacing_mult = 350, 
                save = False, savepath = '', format ='png'):

        '''
        path: recording path, 
        probe: probe (e.g., 'probeA', 'probeB', 'probeC')
        data: 3D array (trials x samples x channels)
        
        stim_times: list of stimulation times
        pre: pre window in ms
        post: post window in ms
        
        first_ch: first channel to plot
        last_ch: last channel to plot

        probeID: probe name ('A', 'B', 'C')
        spike_overlay: whether to overlay spike times from dataframe
        units: dataframe with spike times
        title: title of the plot

        n_trials: number of trials to plot (bc each trial is overlaid)
        spacing_mult: multiplier for spacing between channels
        save: whether to save the plot
        savepath: where to save the plot
        format: format of the saved plot (png, eps, etc)
        '''
        data_stream = load_datastream(path, probe)
        sample_rate = data_stream.metadata['sample_rate']
        data = get_chunk(path, probe, stim_times, 
                             pre = pre, post = post, 
                             chs =np.arange(0,300,1), 
                             median_subtraction = median_subtraction)
        

        probeID = probe.strip('probe')
        total_samps = int((pre/1000 * sample_rate) + (post/1000 * sample_rate))            
        
        if spike_overlay == True:
            stim_indices = np.searchsorted(data_stream.timestamps,stim_times)
            condition = (
                (units['ch'] >= first_ch) &
                (units['ch'] <= last_ch) &
                (units['probe'] == probeID) &
                (units['group'] == 'good')
            )
        
            spikes = np.array(units.loc[condition, 'spike_times'])
            spike_ch = np.array(units.loc[condition, 'ch'])

            spike_dict = {}
            for i, stim in enumerate(stim_indices):
                start_index = int(stim - ((pre/1000)*sample_rate))
                end_index = int(stim + ((post/1000)*sample_rate))  
                window = data_stream.timestamps[start_index:end_index]
                filtered_spikes = [spike_times[(spike_times >= window[0]) & (spike_times <= window[-1])] for spike_times in spikes]  
                spike_dict[i] = filtered_spikes

        ## plotting 
        
        trial_subset = np.linspace(0,len(stim_times)-1, n_trials) #choostrse random subset of trials to plot 
        trial_subset = trial_subset.astype(int)
        #set color maps
        cmap = sns.color_palette("crest",n_colors = n_trials)
        #cmap = sns.cubehelix_palette(n_trials)
        colors = cmap.as_hex()
        if spike_overlay == True:
            cmap2 = sns.color_palette("ch:s=.25,rot=-.25", n_colors = len(spikes))
            colors2 = cmap2.as_hex()
        fig=plt.figure(figsize=(16,24))
        time_window = np.linspace(-pre,post,(total_samps))
        for trial,color in zip(trial_subset,colors):
        
            for ch in range(first_ch,last_ch): 
                plt.plot(time_window,data[trial,:,ch]+ch*spacing_mult,color=color)
        
            if spike_overlay == True:
                for i,ch in enumerate(spike_ch): 

                    if spike_dict[trial][i].size > 0:
                        for spike in spike_dict[trial][i]:
                            spike = spike - stim_times[trial]
                            plt.scatter(spike*1000, (spike/spike) + ch*spacing_mult, 
                            alpha = 0.75, color = colors2[i], s = 500)
        
        plt.gca().axvline(0,ls='--',color='r')       
        plt.xlabel('time from stimulus onset (ms)')
        plt.ylabel('uV')
        plt.title(title)
        
        if save == True:
            plt.gcf().savefig(savepath,format=format,dpi=600)


import numpy as np
import pandas as pd
import warnings
import os, h5py, csv
from glob import glob
from dlab.generalephys import option234_positions
from dlab.sglx_analysis import readAPMeta
from dlab.utils import get_peak_waveform_from_template
import dlab.continuous_traces as ct
from random import sample

try:
	from nwb.nwb import NWB
	from nwb.nwbts import TimeSeries
except:
	print('no Allen Institute NWB API. get this from http://stash.corp.alleninstitute.org/projects/INF/repos/ainwb/browse')
try:
	from djd.OpenEphys import loadContinuous, loadFolder, load_kwik_klusters
except:
	try:
		from OpenEphys import loadContinuous, loadFolder, load_kwik_klusters
	except:
		print('no OpenEphys.py get this from https://github.com/open-ephys/analysis-tools')

warnings.simplefilter(action='ignore', category=FutureWarning)
		
def load_phy_template(path,cluster_file='KS2',site_positions = option234_positions, **kwargs):
	"""load spike data that has been manually sorted with the phy-template GUI

    Parameters
    ----------
    path : string
        the path to the sorted data
    cluster_file : string, optional
        the format of the cluter_info file. options, KS2, KS3
    site_positions : np.array, optional
        the geometry of the sites on the array. n x 2, where n is the number of channels. the site_positions should contain coordinates of the channels in probe space. for example, in um on the face of the probe

    Returns
    -------
    dict
		returns a dictionary of 'good' units, each of which includes:
			times: spike times, in seconds
			template: template used for matching
			ypos: y position on the probe, calculated from the template. requires an accurate site_positions. averages template from 100 spikes.
			xpos: x position on the probe, calcualted from the template. requires an accurate site_positions. averages template from 100 spikes.
	""" 
	clusters = np.load(open(os.path.join(path,'spike_clusters.npy'),'rb'))
	spikes = np.load(open(os.path.join(path,'spike_times.npy'),'rb'))
	spike_templates = np.load(open(os.path.join(path,'spike_templates.npy'),'rb'))
	templates = np.load(open(os.path.join(path,'templates.npy'),'rb'))
	cluster_id,KSlabel,KSamplitude,KScontamination = [],[],[],[]
	[KSlabel.append(row) for row in csv.reader(open(os.path.join(path,'cluster_KSLabel.tsv')))]
	[KSamplitude.append(row) for row in csv.reader(open(os.path.join(path,'cluster_Amplitude.tsv')))]
	[KScontamination.append(row) for row in csv.reader(open(os.path.join(path,'cluster_ContamPct.tsv')))]
	if os.path.isfile(os.path.join(path,'cluster_group.tsv')):
		# cluster_id = [row for row in csv.reader(open(os.path.join(path,'cluster_group.tsv')))][1:]
		[cluster_id.append(row) for row in csv.reader(open(os.path.join(path,'cluster_group.tsv')))]
	else:
		if os.path.isfile(os.path.join(path,'cluster_groups.csv')):
			# cluster_id = [row for row in csv.reader(open(os.path.join(path,'cluster_groups.csv')))][1:]
			[cluster_id.append(row) for row in csv.reader(open(os.path.join(path,'cluster_groups.csv')))]
		else: print('cant find cluster groups, either .tsv or .csv')
	if 'sampling_rate' in kwargs.keys():
		samplingrate = kwargs['sampling_rate']
	else:
		samplingrate =30000.
		# print('no sampling rate specified, using default of 30kHz')
		
	units = {}
	for i in np.arange(1,np.shape(cluster_id)[0]):
		unit = int(cluster_id[i][0].split('\t')[0])
		units[str(unit)] = {}
		
		#get the unit spike times
		units[str(unit)]['samples'] = spikes[np.where(clusters==unit)].flatten()
		units[str(unit)]['times'] = spikes[np.where(clusters==unit)]/samplingrate
		units[str(unit)]['times'] = units[str(unit)]['times'].flatten()
		
		#get the mean template used for this unit
		all_templates = spike_templates[np.where(clusters==unit)].flatten()
		n_templates_to_subsample = 100
		random_subsample_of_templates = templates[all_templates[np.array(np.random.rand(n_templates_to_subsample)*all_templates.shape[0]).astype(int)]]
		mean_template = np.mean(random_subsample_of_templates,axis=0)
		units[str(unit)]['template'] = mean_template
		
		#take a weighted average of the site_positions, where the weights is the absolute value of the template for that channel
		#this gets us the x and y positions of the unit on the probe.
		# print(mean_template.T.shape)
		weights = np.zeros(site_positions.shape)
		for channel in range(mean_template.T.shape[0]):
			weights[channel,:]=np.trapz(np.abs(mean_template.T[channel,:]))
		weights = weights/np.max(weights)
		low_values_indices = weights < 0.25  # Where values are low,
		weights[low_values_indices] = 0      # make the weight 0
		(xpos,ypos)=np.average(site_positions,axis=0,weights=weights)
		units[str(unit)]['waveform_weights'] = weights
		units[str(unit)]['xpos'] = xpos
		units[str(unit)]['ypos'] = ypos #- site_positions[-1][1]
		units[str(unit)]['label'] =  cluster_id[i][0].split('\t')[1]
		units[str(unit)]['KSlabel'] =  KSlabel[i][0].split('\t')[1]
		units[str(unit)]['KSamplitude'] =  KSamplitude[i][0].split('\t')[1]
		units[str(unit)]['KScontamination'] =  KScontamination[i][0].split('\t')[1]
	return units

def df_from_phy(path,site_positions = option234_positions,**kwargs):
	"""load spike data that has been manually sorted with the phy-template GUI

    Parameters
    ----------
    path : string
        the path to the sorted data
    cluster_file : string, optional
        the format of the cluter_info file. options, KS2, KS3
    site_positions : np.array, optional
        the geometry of the sites on the array. n x 2, where n is the number of channels. the site_positions should contain coordinates of the channels in probe space. for example, in um on the face of the probe

    Returns
    -------
    pandas DataFrame
		returns a DataFrame of 'good' units, each of which includes:
	"""
	nwb_data = load_phy_template(path,site_positions)
	#structures is a dictionary that defines the bounds of the structure e.g.:{'v1':(0,850), 'hpc':(850,2000)}
	mouse = [];experiment=[];cell = [];ypos = [];xpos = [];waveform=[];template=[];structure=[];times=[]
	index = []; count = 1
	nwb_id = [];probe_id=[]
	depth=[];#print(list(nwb_data.keys()));print(list(nwb_data['processing'].keys()));
	if 'probe' in kwargs.keys():
		for probe in list(nwb_data['processing'].keys()):
			if 'UnitTimes' in list(nwb_data['processing'][probe].keys()): 
				for i,u in enumerate(list(nwb_data['processing'][probe]['UnitTimes'].keys())):
					if u != 'unit_list':
						nwb_id.append(nwbid)
						probe_id.append(probe)
						index.append(count);count+=1
						mouse.append(str(np.array(nwb_data.get('identifier'))))
						experiment.append(1)
						cell.append(u)
						times.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['times']));# print(list(nwb_data['processing'][probe]['UnitTimes'][u].keys()))
						if 'ypos' in list(nwb_data['processing'][probe]['UnitTimes'][u].keys()):
							ypos.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['ypos']))
							has_ypos = True
						else:
							ypos.append(None)
							has_ypos = False				
						if 'depth' in list(nwb_data['processing'][probe]['UnitTimes'][u].keys()):
							depth.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['depth']))
						else:
							if has_ypos:
								depth.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['ypos']))
							else:
								depth.append(None)
						if 'xpos' in list(nwb_data['processing'][probe]['UnitTimes'][u].keys()):
							xpos.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['xpos']))
							has_xpos = True
						else:
							xpos.append(None)
							has_xpos = False
						template.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['template']))
						waveform.append(get_peak_waveform_from_template(template[-1]))
						if not structures == None:
							structur = None
							for struct, bounds in structures.iteritems():
								if ypos[-1] > bounds[0] and ypos[-1]< bounds[1] :
									structur=struct
						else:
							structur = None
						structure.append(structur)
	df = pd.DataFrame(index=index)
	df = df.fillna(np.nan)
	df['nwb_id'] = nwb_id
	df['mouse'] = mouse
	df['experiment'] = experiment
	df['probe'] = probe_id
	df['structure'] = structure
	df['cell'] = cell
	df['times'] = times
	df['ypos'] = ypos
	df['xpos'] = xpos
	df['depth'] = depth
	df['waveform'] = waveform
	df['template'] = template
	return df

class UnitData:
    def __init__(self,recording_path) -> None:
        self.old_cwd       = os.getcwd()
        self.sampling_rate = 30000.0
        
        if 'recording' not in os.path.basename(recording_path):
            print('Please provide a path to a recording folder (e.g. /path/to/recording1)')
            
        self.recording_path = recording_path
        self.all_folders    = glob(os.path.join(self.recording_path,'continuous','*'))
        self.ap_folders     = [folder for folder in self.all_folders if 'LFP' not in os.path.basename(folder)]
        self.ap_folders     = [folder for folder in self.ap_folders if 'NI-DAQmx' not in os.path.basename(folder)]
        
    def load(self,probe_depths=[],probes=[],acq='OpenEphys',ignore_phy=False):
        # site_positions = self.option234_positions
        
        for i,PROBE in enumerate(probes):
            print(PROBE)
            probe_name = PROBE
            for folder in self.ap_folders:
                if 'Probe'+PROBE in folder:
                    probe_path = folder
            
            os.chdir(probe_path)
            
            if os.path.isfile('spike_seconds.npy'):
                spike_times = np.load('spike_seconds.npy')
            else:
                spike_times = np.load('spike_times.npy').flatten()
                if isinstance(spike_times[0],(np.uint64)):
                    if acq == 'OpenEphys':
                        try:
                            ts = np.load('timestamps.npy')
                            spike_times = ts[spike_times]
                            np.save('spike_seconds.npy',spike_times)
                        
                        except: 
                            print('could not load timestamps.npy')
                            spike_times = spike_times/self.sampling_rate
                            np.save('spike_seconds.npy',spike_times)
                            

                    else: print('SpikeGLX currently not supported')
                    
                    
            site_positions  = np.load('channel_positions.npy')               
            spike_clusters  = np.load('spike_clusters.npy').flatten()
            spike_templates = np.load('spike_templates.npy')
            templates       = np.load('templates.npy')
            amplitudes      = np.load('amplitudes.npy')
            
            cluster_info = None        
            try:
                cluster_info = pd.read_csv('cluster_info.tsv', delimiter='\t')
                if ignore_phy == True:
                    cluster_info    = None
                    cluster_Amps    = pd.read_csv('cluster_Amplitude.tsv', delimiter='\t')
                    ContamPct       = pd.read_csv('cluster_ContamPct.tsv', delimiter='\t')
                    KSLabel         = pd.read_csv('cluster_KSLabel.tsv', delimiter='\t')
            except: 
                print('Unable to load cluster_info.tsv. Have you opened this data in Phy?')
                cluster_Amps    = pd.read_csv('cluster_Amplitude.tsv', delimiter='\t')
                ContamPct       = pd.read_csv('cluster_ContamPct.tsv', delimiter='\t')
                KSLabel         = pd.read_csv('cluster_KSLabel.tsv', delimiter='\t')

            
            weights        = np.zeros(site_positions.shape)
            mean_templates = []
            peak_templates = []

            all_weights    = []
            amps           = []
            times          = []
            ch             = []
            
            
            for unit_id in np.unique(spike_clusters):
                #get mean template for each unit
                all_templates,count    = np.unique(spike_templates[np.where(spike_clusters==unit_id)],return_counts=True)
                
                if len(all_templates) > 100:
                    n_templates_to_subsample = 100
                else: 
                    n_templates_to_subsample = len(all_templates)
                
                random_subsample_of_templates = templates[sample(list(all_templates),n_templates_to_subsample)]
                
                mean_template = np.mean(random_subsample_of_templates,axis=0)
                
                mean_templates.append(mean_template)
                
                if cluster_info is not None:
                    best_ch = cluster_info[cluster_info.cluster_id == unit_id].ch.values[0].astype(int)
                else:
                    best_ch = np.argmax((np.max(mean_template,axis=0) - np.min(mean_template,axis=0)))
                
                ch.append(best_ch)
                
                peak_wv = mean_template[:,best_ch-1]
                peak_templates.append(peak_wv)
                
                #Take weighted average of site positions where hweights is abs value of template for that channel
                #This gets us the x and y positions of the unit on the probe

                # for channel in range(len(mean_template.T)):
                #     weights[channel,:] = np.trapz(np.abs(mean_template.T[channel]))
            
                # # weights                /= weights.max()
                # weights[weights < 0.25] = 0 #Where weights are low, set to 0
                # x,y                     = np.average(site_positions,weights=weights,axis=0)
                # all_weights.append(weights)
                # xpos.append(x)
                # ypos.append(y)
                
                amps.append(amplitudes[:,0][spike_clusters==unit_id])
                times.append(spike_times[spike_clusters==unit_id])

            if cluster_info is not None:
                probe_data = cluster_info.copy()
            else:
                probe_data = pd.DataFrame()
                probe_data['cluster_id'] = np.unique(spike_clusters)
                probe_data['Amplitude']  = cluster_Amps[np.in1d(cluster_Amps.cluster_id.values,np.unique(spike_clusters))]['Amplitude'].values 
                probe_data['ContamPct']  = ContamPct[np.in1d(ContamPct.cluster_id.values,np.unique(spike_clusters))]['ContamPct'].values 
                probe_data['KSLabel']    = KSLabel[np.in1d(KSLabel.cluster_id.values,np.unique(spike_clusters))]['KSLabel'].values 
                probe_data['ch']         = ch
                
            probe_data['probe']      = [probe_name]*len(probe_data)
            # probe_data['shank']    = np.floor(cluster_info['xcoords'].values / 205.).astype(int)
            probe_data['depth']      = np.array(site_positions[:,1][ch])*-1 + probe_depths[i]
            probe_data['times']      = times
            probe_data['amplitudes'] = amps
            probe_data['template']   = mean_templates
            # probe_data['weights']  = all_weights
            probe_data['peak_wv']    = peak_templates
            probe_data['xpos']       = site_positions[:,0][probe_data['ch']]
            probe_data['ypos']       = site_positions[:,1][probe_data['ch']]
            probe_data['n_spikes']   = [len(i) for i in times]
            
            if 'unit_data' not in locals():
                unit_data = probe_data
                
            else:
                unit_data = pd.concat([unit_data,probe_data],ignore_index=True)
                unit_data.reset_index(inplace=True,drop=True)
                
            os.chdir(self.old_cwd)
        
        return unit_data
    
    def get_qMetrics(self,path):
        metrics_path = os.path.join(path,'qMetrics')
        if not os.path.isdir(metrics_path):
            print('Please provide path containing qMetrics folder')
        else:
            params = pd.read_parquet(os.path.join(metrics_path,'_bc_parameters._bc_qMetrics.parquet'))
            
            qMetrics = pd.read_parquet(os.path.join(metrics_path,'templates._bc_qMetrics.parquet'))
            
        return params, qMetrics
    
    def qMetrics_labels(self, probes=['A','B','C'],param_changes = {}):
        ids = []
        all_labels = []            
        for i,PROBE in enumerate(probes):
            labels = []
            for folder in self.ap_folders:
                if 'Probe'+PROBE in folder:
                    probe_path = folder
            
            os.chdir(probe_path)
            
            param, qMetric = self.get_qMetrics(probe_path)
            cluster_id     = qMetric.phy_clusterID.values.astype(int)
            unit_type      = np.full((len(qMetric)),np.nan)
            
            if param_changes:
                for key in param_changes.keys():
                    param[key] = param_changes[key]
            
            # Noise Cluster Condtions
            noise0  = pd.isnull(qMetric.nPeaks)
            noise1  = qMetric.nPeaks                      > param.maxNPeaks[0]
            noise2  = qMetric.nTroughs                    > param.maxNTroughs[0]
            noise3  = qMetric.spatialDecaySlope           > param.minSpatialDecaySlope[0]
            noise4  = qMetric.waveformDuration_peakTrough < param.minWvDuration[0]
            noise5  = qMetric.waveformDuration_peakTrough > param.maxWvDuration[0]
            noise6  = qMetric.waveformBaselineFlatness    > param.maxWvBaselineFraction[0]
            
            unit_type[noise0 | noise1 | noise2 | noise3 | noise4 | noise5 | noise6] = 0 #NOISE
            
            #MUA Conditions
            mua0 = qMetric.percentageSpikesMissing_gaussian > param.maxPercSpikesMissing[0]
            mua1 = qMetric.nSpikes                          < param.minNumSpikes[0]
            mua2 = qMetric.fractionRPVs_estimatedTauR       > param.maxRPVviolations[0]
            mua3 = qMetric.presenceRatio                    < param.minPresenceRatio[0]
            
            unit_type[(mua0| mua1 | mua2 | mua3)&np.isnan(unit_type)] = 2 #MUA
            
            #Optional MUA metrics
            if param.computeDistanceMetrics[0] == 1 & ~param.isoDmin.isna()[0]:
                mua4 = qMetric.isoD   < param.isoDmin[0]
                mua5 = qMetric.Lratio > param.lratioMax[0]
                
                unit_type[(mua4 | mua5)&np.isnan(unit_type)] = 2 #MUA

            else:
                print('No distance metrics calculated')
                
            if param.extractRaw[0] == 1:
                mua6 = qMetric.rawAmplitude       < param.minAmplitude[0]
                mua7 = qMetric.signalToNoiseRatio < param.minSNR[0]
                unit_type[(mua6 | mua7)&np.isnan(unit_type)] = 2 #MUA

            else:
                print('Raw waveforms not extracted')
                                    
            # Somatic Cluster Conditions
            if param.splitGoodAndMua_NonSomatic[0]:
                nsom0 = qMetric.isSomatic != param.somatic[0]
                unit_type[(unit_type==1) & (nsom0)] = 3 #Good Non-Somatic
                unit_type[(unit_type==2) & (nsom0)] = 4 #MUA Non-Somatic
            
            #GOOD Conditions
            unit_type[np.isnan(unit_type)] = 1 #Good

            for i in unit_type:
                if i == 0:
                    labels.append('NOISE')
                    all_labels.append('NOISE')
                if i == 1:
                    labels.append('GOOD')
                    all_labels.append('GOOD')
                if i == 2:
                    labels.append('MUA')
                    all_labels.append('MUA')
                if i == 3:
                    labels.append('NON-SOMA GOOD')
                    all_labels.append('NON-SOMA GOOD')
                if i == 4:
                    labels.append('NON-SOMA MUA')
                    all_labels.append('NON-SOMA MUA')
                    
            ids += list(cluster_id)
          
            out_df = pd.DataFrame({'cluster_id':cluster_id, 'bc_unitType':labels})
        # if os.path.isfile('cluster_bc_unitType.tsv'):
        #     q = input('Would you like to overwrite cluster_bc_unitType.tsv? (Y/N)')
        #     if q == 'Y':
        #         print('Overwriting  cluster_bc_unitType.tsv')
        #         out_df.to_csv('cluster_bc_unitType.tsv', sep='\t', index=False, header=True)
        #     if q == 'N':
        #         print('No output saved')
            print('Saving output....')
            out_df.to_csv('cluster_bc_unitType.tsv', sep='\t', index=False, header=True)
                
        return {'cluster_id':ids, 'qm_labels':all_labels}



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 18 14:36:41 2014

@author: djd
"""
import numpy as np

def warpFileToMatrix(warp,plot=False):
    for i in range(np.shape(warp)[0]):
        if warp[i][1] < warp[i+1][1]:
            numCols = i+1
            break
    numRows = (np.shape(warp)[0])/numCols
    X=np.zeros((numRows,numCols))
    Y=np.zeros((numRows,numCols))
    U=np.zeros((numRows,numCols))
    V=np.zeros((numRows,numCols))
    Intensity=np.zeros((numRows,numCols))
    for i in range(np.shape(warp)[0]):
        row = numRows - 1 - np.floor(i/numCols)
        col = i % numCols
        X[row][col]=warp[i][0]
        Y[row][col]=warp[i][1]
        U[row][col]=warp[i][2]
        V[row][col]=warp[i][3]
        Intensity[row][col]=warp[i][4]
    if plot:
        plt.subplot(151);plt.imshow(X)
        plt.subplot(152);plt.imshow(Y)
        plt.subplot(153);plt.imshow(U)
        plt.subplot(154);plt.imshow(V)
        plt.subplot(155);plt.imshow(Intensity)
    return X,Y,U,V,Intensity

def changeWarpAspect(warp,x=1):
    warpout = np.zeros((0,5))#make an empty first row of correct shape
    for i in range(np.shape(warp)[0]):
        if warp[i][0] < x and warp[i][0] > -1*x:
            warpout = np.vstack((warpout,warp[i,:]))
    np.delete(warpout,0,0)#get rid of the empty first row
    return warpout

def matrixToWarpFile(X,Y,U,V,I):
    numCols = np.shape(X)[1];print numCols
    numRows = np.shape(X)[0];print numRows
    warp = np.zeros((numCols*numRows,5))
    for i in range(np.shape(warp)[0]):
        row = int(np.floor(i/numCols))
        col = i % numCols
        print (row,col)
        warp[i][0]=X[row][col]
        warp[i][1]=Y[row][col]
        warp[i][2]=U[row][col]
        warp[i][3]=V[row][col]
        warp[i][4]=I[row][col]
    return warp

##strecth the Y warp only in specified xRange. default is the whole xRange
#def warpStretchY(warp,amount,area=(0,1)):
# warpout = warp
# xAspect = abs(warp[0][0])
# for i in range(np.shape(warp)[0]):
# if abs(warp[i][0]) <= area[1]*xAspect and abs(warp[i][0]) >= area[0]*xAspect :
# warpout[i][3]=warp[i][3]*amount #stretch warp
# return warpout
#
##strecth the X warp only in specified yRange. default is the whole yRange
#def warpStretchX(warp,amount,area=(0,1)):
# warpout = warp
# yAspect = abs(warp[0][1])
# for i in range(np.shape(warp)[0]):
# if abs(warp[i][1]) < area[1]*yAspect and abs(warp[i][1]) > area[0]*yAspect :
# warpout[i][2]=warp[i][2]*amount #stretch warp
# return warpout

#strecth the Y warp only in specified xRange. default is the whole xRange
def warpStretchY(warp,amount,area=(0,1)):
    warpout = warp
    xAspect = abs(warp[0][0])
    for i in range(np.shape(warp)[0]):
        if abs(warp[i][0]) <= area[1]*xAspect and abs(warp[i][0]) >= area[0]*xAspect :
            warpout[i][3]=warp[i][3]*amount #stretch warp
    return warpout
#strecth the X warp only in specified yRange. default is the whole yRange
def warpStretchX(warp,amount,area=(0,1)):
    warpout = warp
    yAspect = abs(warp[0][1])
    for i in range(np.shape(warp)[0]):
            if abs(warp[i][1]) < area[1]*yAspect and abs(warp[i][1]) > area[0]*yAspect :
                warpout[i][2]=warp[i][2]*amount #stretch warp
                return warpout

#strecth the warp only in specified yRange. default is the whole yRange
def warpStretch(warp,amount=0.1,limits=(0,0)):
# warpout = warp
# for i in range(np.shape(warp)[0]):
# if abs(warp[i][0]) > limits[0] and warp[i][1] > limits[1] :
# scale = ((abs(warp[i][0])*(1/abs(warp[0][0])))*amount)
# warpout[i][1]=(1+scale)*warp[i][1] #stretch warp
# return warpout
    X,Y,U,V,I = warpFileToMatrix(warp,plot=False)
    #newY = (Y+0.5) * np.ceil(I);newY[0,0]=Y[0,0];newY[Y.shape[0]-1,0]=Y[Y.shape[0]-1,0];
    newY = (Y) * np.ceil(I)   
    
    middleCol = int(np.shape(Y)[1]/2)
    middleRow = np.nonzero(newY)[0][0]+(np.nonzero(newY)[0][len(np.nonzero(newY)[0])-1] - np.nonzero(newY)[0][0])/2
    for i in range(middleCol):
        columnL = middleCol-i-1;
        columnR = middleCol+i
        bottom = middleRow+5
        if sum(np.abs(newY[:,columnL]))>0 and columnL>0:
            top = np.nonzero(newY[:,columnL])[0][0];#print (top,bottom)
            newTop = top + (i/2)
            newY[top:bottom,columnL]= np.linspace(Y[newTop,columnL],Y[bottom,columnL],(bottom-top))
            newY[top:bottom,columnR]= np.linspace(Y[newTop,columnR],Y[bottom,columnR],(bottom-top))
    for i in range(np.shape(newY)[0]):
        for j in range(np.shape(newY)[1]):
            if I[i,j]==0:
                newY[i,j]=Y[i,j]
                
#    for i in range(np.shape(newY)[1]):
#        if sum(np.abs(newY[:,i]))>0 and i>0:
#             if (np.abs(50-i)) > 12:
#                 for j in range(np.shape(newY)[0]-1):
#                     if newY[j,i] == 0 and newY[j+1,i] != 0:
#                        top = j+1
#                     if newY[j,i] != 0 and newY[j+1,i] == 0 or j == len(newY[:,i])-1:
#                        bottom = j+1
#                 stretch = round(((bottom-top)-13)/2)
#                 print (54-top)
#                 newY[top:bottom,i]=np.linspace(newY[47+((57-top)/2),i],newY[bottom,i],num=bottom-top)
##                    print 59+(np.abs(24-np.abs(50-i)))
#            #newY[:,i] = newY[:,i]*(abs(50-i)*amount+1)
#            #newY[:,i] =  newY[:,i]*(np.abs(50-i)/10+1)
    return newY


import numpy as np
import random
from scipy import stats
from scipy.signal.windows import boxcar
from scipy.signal import convolve,correlate,resample,argrelextrema
from scipy.cluster.vq import kmeans,kmeans2
from scipy.stats import pearsonr
from dlab import cleanAxes
from dlab import psth_and_raster as psth_

def smooth_boxcar(data,boxcar_size):
    """smooths an impulse respone of an already computed receptive field. uses a boxcar to smooth.

    Parameters
    ----------
    data : np.array
        the 1d temporal kernel to smooth
    size : int, optional
        the width of the boxcar used to smooth (default is
        3)

    Returns
    -------
    np.array
        the smoothed input kernel
    """    
    smoothed = convolve(data,boxcar(int(boxcar_size)))/boxcar_size
    smoothed = smoothed[int(boxcar_size/2):len(data)+int(boxcar_size/2)]
    return smoothed

def moving_average(a, n=3) :
    """calculates a moving average of the input data with widnow 

    Parameters
    ----------
    a : np.array
        the 1d temporal input to average
    size : int, optional
        the width of window to average within (default is 3)

    Returns
    -------
    np.array
        average, same shape as input a
    """    
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n


    return peak

def precision(spike_times,stimulus_times,boxcar_size = 5,precision_type='first',pre=0.,post=15.,binsize=0.01,threshold=0.05,find_events=True):
    """precision - calculate the temporal jitter of first spike, or jitter of total spikes, depending on type flag

    Parameters
    ----------
    spike_times : np.array
        the times of events (usually action potentials from one neuron) to measure the precision of
    stimulus_times : np.array
        the times of events (often stimuli) that are repeated to measure precision around
    boxcar_size : int, optional
        the width of the boxcar used to smooth (default is 5). passed to smooth_boxcar
    precision_type: str, optional
        default : 'first'. options : 'first', anything else uses all spikes
    pre=0.
        the time before each entry in stimulus_times to use
    post=15.
        the time after each entry in stimulus_times to use
    binsize=0.01
        the binsize to use when finding events
    threshold : float, default = 0.05
        the threshold for finding events
    find_events : bool, default = True
        whether or not to find events
    
    Returns
    -------
    tuple
        (np.mean(event_precision),np.std(event_precision),event_precision)    
    """        
 
    if find_events:
        smoothed,all_events,events = get_events(spike_times,stimulus_times,boxcar_size = boxcar_size,threshold=threshold,pre=pre,post=post,binsize=binsize)
    else:
        events = [0.]
    #print events
    #get spikes for each event
    event_precision = []
    for i,event in enumerate(events):
        all_spikes = []
        first_spikes = []
        for trial in stimulus_times:
            event_start = trial + event
            if i == len(events):
                if event[i+1] - event > 0.05:
                    event_end = trial + event[i+1]
                else:
                    event_end = trial + event + 0.05
            else:
                event_end = trial + event + 0.05
            indices = np.where((spike_times > event_start) & (spike_times < event_end))[0] 

            first_spike = spike_times[np.where(spike_times>event_start)[0][0]]-event_start
            #find jitter of first spikes
            if first_spike < .1: #arbitrary cutoff because we don't trust identified events that don't have a spike within 100 msec
                first_spikes.extend([first_spike])

            #find jitter of all spikes
            event_spikes = np.array(spike_times[indices])-event_start
            all_spikes.extend(event_spikes)
  
        if precision_type =='first':
            #event_precision.extend([np.median(np.array(first_spikes))])
            #event_precision.extend([np.median(np.array(first_spikes)-np.min(first_spikes))])
            all_event_spikes = np.sort(np.array(first_spikes))
        else:
            all_event_spikes = np.sort(np.array(all_spikes).flatten())
            
        #print all_event_spikes
        first_half = all_event_spikes[:np.where(all_event_spikes<np.median(all_event_spikes))[0][-1]]
        second_half = all_event_spikes[np.where(all_event_spikes>np.median(all_event_spikes))[0][0]:]
        event_precision.extend([np.median(second_half)-np.median(first_half)])
    
    event_precision = np.array(event_precision)
    event_precision = event_precision[~np.isnan(event_precision)] * 1000.
    
    return (np.mean(event_precision),np.std(event_precision),event_precision)

def get_binarized(spike_times,stimulus_times,pre=0.,post=15.,convolve=0.):
    """create a 2D array, trials (rows) x time (column), where a 1 indicates an event (spike) on that trial in that time. uses 1 msec time bins. 
    
    Parameters
    ----------
    spike_times : np.array
        the times of events (usually action potentials from one neuron) to measure
    stimulus_times : np.array
        the times of events (often stimuli) that are repeated (i.e., trial times)
    pre : float, default = 0.
        the time before each entry in stimulus_times to use
    post : float, default = 15.
        the time after each entry in stimulus_times to use
    convolve : int, default = boxcar 
        the binsize to use when finding events

    Returns
    ----------
    binarized : np.array
        trials (rows) x time (column), where a 1 indicates an event (spike) on that trial in that time
    '"""    

    bytrial = psth_.raster(spike_times,stimulus_times,pre=pre,post=post,timeDomain=True,output='data')
    binarized = []
    for trial in bytrial:
        binarized_ = np.zeros(int((post-pre)*1000))#use 1-msec bins
        for spike in trial:
            if spike > pre and spike < post:
                binarized_[int(np.floor(spike*1000))] = 1
        if convolve > 0.001 :
            binarized_=smooth_boxcar(binarized_,convolve)
        binarized.append(binarized_)   
    return binarized

def get_binned(spike_times,stimulus_times,binsize,pre=0.,post=15.,convolve=0.):
    """create a 2D array, trials (rows) x time (column), where the number of events in spike times that occur in each bin are counted. 
    
    Parameters
    ----------
    spike_times : np.array
        the times of events (usually action potentials from one neuron) to measure
    stimulus_times : np.array
        the times of events (often stimuli) that are repeated (i.e., trial times)
    pre : float, default = 0.
        the time before each entry in stimulus_times to use
    post : float, default = 15.
        the time after each entry in stimulus_times to use
    convolve : int, default = boxcar 
        the binsize to use when finding events

    Returns
    ----------
    binarized : np.array [note that this is poorly named. TODO: name it better]
        trials (rows) x time (column), where the value indicates the total events on that trial in that time
    '"""   
    bytrial = psth_.raster(spike_times,stimulus_times,pre=pre,post=post,timeDomain=True,output='data')
    # print(np.array(bytrial).shape())
    binarized = []
    for trial in bytrial:
        binarized_ = np.zeros(int((post-pre)*1000/(1000*binsize)+1))#use binsize msec bins
        for spike in trial:
            if spike > pre and spike < post:
                binarized_[int(np.floor((spike-pre)*1000/(1000*binsize)))] += 1
        if convolve > 0.001 :
            binarized_=smooth_boxcar(binarized_,convolve)
        binarized.append(binarized_)   
    return binarized

def reliability(spike_times,stimulus_times,binsize,pre=0.,post=15.):
    """# how reproducible a spike train is over trials, at the msec level. equivalent to cosine similarity. 
    
    Parameters
    ----------
    spike_times : np.array
        the times of events (usually action potentials from one neuron) to measure
    stimulus_times : np.array
        the times of events (often stimuli) that are repeated (i.e., trial times)
    pre : float, default = 0.
        the time before each entry in stimulus_times to use
    post : float, default = 15.
        the time after each entry in stimulus_times to use

    Returns
    ----------
     : float
        reliability. bounded 0 to 1, where 1 is repeated exactly in the same bin and only those bins across all trials; 0 is no repeated bins across trials.  
    '"""      

    binarized=get_binned(spike_times,stimulus_times,binsize,pre=pre,post=post)
    sum = 0
    for c,i in enumerate(binarized):
        for c2 in np.arange(c+1,np.shape(binarized)[0],1):
            j = binarized[c2][:]
            if np.sum(i) > 0 and np.sum(j) > 0:
                sum += np.inner(i,j) / np.inner(np.linalg.norm(i),np.linalg.norm(j))
            else:
                sum+=0

    #return (2 / float(np.shape(binarized)[0]) * (np.shape(binarized)[0] - 1)) * sum,binarized
    return sum / (np.shape(binarized)[0]/2.)#,binarized

def fano(spike_times,stimulus_times,pre=0.,post=15.,binsize=0.01,boxcar_size = 5,counting_window=0.3,threshold=0.2,by_event=False):
    """measure the fano factor of an input (spike_times) relative to repeats (stimulus times)
    
    Parameters
    ----------
    spike_times : np.array
        the times of events (usually action potentials from one neuron) to measure
    stimulus_times : np.array
        the times of events (often stimuli) that are repeated (i.e., trial times)
    pre : float, default = 0.
        the time before each entry in stimulus_times to use
    post : float, default = 15.
        the time after each entry in stimulus_times to use
    binsize=0.01
        the binsize to use when finding events        
    boxcar_size : int, optional
        the width of the boxcar used to smooth (default is 5). passed to smooth_boxcar
    counting_window : float, optional
        the size of the window (bin) to count spikes in for calculating fano factor
    threshold : float, optional 
        the threshold for finding events. default = 0.2
    by_event : bool, optional
        whether or not to find the fano for all (False) or separately for events that repeat within each trial (True). default = False

    Returns
    ----------
    tuple
        (median fano factor, all spike counts, all fano factors)
    '"""  
    if by_event:
        smoothed,all_events,events = get_events(spike_times,stimulus_times,boxcar_size = boxcar_size,threshold=threshold,pre=pre,post=post,binsize=binsize)
    else:
        events = [0]
    fanos=[]
    for i,event in enumerate(events):
        counts=[]
        for trial in stimulus_times:
            event_start = trial + event
            event_end = trial + event + counting_window
            indices = np.where((spike_times > event_start) & (spike_times < event_end))[0] 
            
            counts.append(len(indices))

        fanos.append(np.std(counts)**2 / np.mean(counts))
    return np.median(fanos),counts, fanos

def get_events(spike_times,stimulus_times,threshold=.05,boxcar_size = 15,pre=0.,post=15.,binsize=0.001):
    """find the events within the response; can find multiple at arbitrary times, based on local minima and maxima in the averaged (around stimulus_times) response
    
    Parameters
    ----------
    spike_times : np.array
        the times of events (usually action potentials from one neuron) to measure
    stimulus_times : np.array
        the times of events (often stimuli) that are repeated (i.e., trial times)
    threshold : float, optional 
        the threshold for finding events. default = 0.05        
    boxcar_size : int, optional
        the width of the boxcar used to smooth (default is 5). passed to smooth_boxcar        
    pre : float, default = 0.
        the time before each entry in stimulus_times to use
    post : float, default = 15.
        the time after each entry in stimulus_times to use
    binsize=0.01
        the binsize to use when finding events        

    Returns
    ----------
    tuple
        (smoothed, event times relative to zero, ? ) TODO: anotate outputs better
    '"""      
    (edges,psth,variance) =  psth_.psth_line(spike_times,
                                             stimulus_times,
                                             pre=pre,post=post,binsize=binsize,output='p',timeDomain=True)
    numbins = int((post-pre) /  binsize)

    # first, find events:
    smoothed = smooth_boxcar(psth[:numbins],int(boxcar_size))
    minima = np.where(np.r_[True, smoothed[2:] > smoothed[:-2]] & np.r_[smoothed[:-2] > smoothed[2:], True]==True)[0]#from http://stackoverflow.com/questions/4624970/finding-local-maxima-minima-with-numpy-in-a-1d-numpy-array
   
    #print np.max(psth)
    threshold = threshold * np.max(psth)
    good_minima = []
    #print threshold
    for i,minimum in enumerate(minima[:-1]):
        if minima[i+1]*binsize - minima[i]*binsize > 0.3:
            num_bins_after_minimum = 0.3 / binsize
        else:
            num_bins_after_minimum = minima[i+1] - minima[i]
        try:
            if np.max(psth[minimum:minimum+num_bins_after_minimum]) > threshold:
                good_minima.extend([minimum])
        except:
            pass
            # print np.shape(psth)
            # print minimum
            # print num_bins_after_minimum
            # print threshold
            # return np.nan,np.nan,np.nan
            # 

    return smoothed,minima*binsize,np.array(good_minima)*binsize - (boxcar_size/2.)*binsize

def entropy(spike_times,stimulus_times,wordlength,binsize=0.001,pre=0.,post=15.):
    """measure the entropy in spike_times relative to stimulus_times
    
    Parameters
    ----------
    spike_times : np.array
        the times of events (usually action potentials from one neuron) to measure
    stimulus_times : np.array
        the times of events (often stimuli) that are repeated (i.e., trial times)
    word_length : int, 
        how many bins are in the words in the entropy dictionary
    binsize=0.01
        the binsize to use when finding events     
    pre : float, default = 0.
        the time before each entry in stimulus_times to use
    post : float, default = 15.
        the time after each entry in stimulus_times to use
   

    Returns
    ----------
    entropy: float
        the entropy of spike_times relative to stimulus_times
    '"""      
    binarized=get_binarized(spike_times,stimulus_times,pre=pre,post=post)

    # create words of length wordlength 
    entropies_per_time=[]
    for t in range(len(binarized[0])):
        words = []
        for trial in binarized:
            if t<len(trial) - wordlength: #cutoff of the end of each trial because there aren't enough bins left to make the word
                word = trial[t:t+wordlength]
                words.append(word)

        #make a distribution of the words
        p = {}
        #find all the words that actually occured. and the frequency of their occurence
        #from http://stackoverflow.com/questions/16970982/find-unique-rows-in-numpy-array
        # and http://stackoverflow.com/questions/33235135/count-occurrences-of-arrays-in-multidimensional-arrays-in-python
        words = np.array(words)
        b = np.ascontiguousarray(words).view(np.dtype((np.void, words.dtype.itemsize * wordlength)))
        _, idx,counts = np.unique(b, return_index=True,return_counts=True)
        possible_words = words[idx]
        p = dict(zip([str(word) for word in possible_words],counts/float(np.shape(np.array(words))[0])))
        print('all the words that occured:'+str(np.shape(np.array(possible_words))[0]),end="\r")

        # calculate entropy as in reinagel and reid 2000; kumbhani et al., 2007
        sum = 0
        for word in possible_words:
            sum +=  p[str(word)] * np.log2(p[str(word)])
        H = (-1 / (wordlength*binsize)) * sum
        entropies_per_time.append(H)
        
    return np.mean(entropies_per_time)

def fano2(a):
# computes the Fano factor , variance**2/mean for an input matrix a with dimensions n x m, 
# where n is the number of trials and m is the number of bins in the response
    return np.std(a,axis=0)**2/np.mean(a,axis=0)

def ccmax(a):
# computes the cc_max for an input matrix a with dimensions n x m, 
# where n is the number of trials and m is the number of bins in the response
    ntrials = a.shape[0]
    corr_matrix = np.empty((ntrials,ntrials))
    for i in range(ntrials):
        for j in range(ntrials):
            r,p = pearsonr(a[i,:],a[j,:])
            corr_matrix[i,j] = r

    inds = np.triu_indices(ntrials, k=1)
    upper = corr_matrix[inds[0],inds[1]]
    return np.nanmean(upper)

def mutual_information(spike_times,stimulus_times,wordlength,binsize=0.001,pre=0.,post=15.,method='poisson'):
    indices = np.where((spike_times > stimulus_times[0]) & (spike_times < stimulus_times[-1]))[0] 
    if method == 'poisson':
        reference_spikes = possion_times(np.array(spike_times)[indices])
    if method == 'shift':
        reference_spikes = shifted_times(np.array(spike_times)[indices],stimulus_times)
    total = entropy(reference_spikes,stimulus_times,wordlength,binsize=0.001,pre=0.,post=15.)
    noise = entropy(np.array(spike_times),stimulus_times,wordlength,binsize=0.001,pre=0.,post=15.)
    return total,noise, total-noise, (total-noise) / (post - pre)

def z_value(spike_times,stimulus_times,binsize,pre=0.,post=15.,method='shift'):
#based on reinagel and reid, 2000
# their definition, sightly modified for easier notation here
# Z(binsize) = limI(L,binsize) - I(L=1,binsize)
# "The term I(L=1,binsize) represents the estimate of information rate that would be obtained on the approximation that time bins are statistically independent within the spike train."
    lim_I = mutual_information(spike_times,stimulus_times,wordlength,binsize=binsize,pre=0.,post=15.,method='poisson')
    I1 = mutual_information(spike_times,stimulus_times,1,binsize=binsize,pre=0.,post=15.,method='poisson')
    return z

def possion_times(spike_times):
#given an input spike train, make the times a rate-matched Poisson process
    rate = len(np.array(spike_times)) / (np.array(spike_times)[-1] - np.array(spike_times)[0])
    t = np.array(spike_times)[0]
    poiss = [t]
    for i in range(len(spike_times)-1):
        t+=random.expovariate(rate)
        poiss.append(t)
    return np.array(poiss)

def shuffled_times(spike_times,window):
#given an input spike train, shuffle the spike times preserving the structure within the window
    return spike_times

def shifted_times(spike_times,stimulus_times):
#given an input spike train and stimulus times, 
#shift the spike times in each interval by an random amount, with wrapping
    shifted_times = np.zeros(len(spike_times))
    for i,start in enumerate(stimulus_times[:-1]):
        indices = np.where((spike_times > start) & (spike_times <= stimulus_times[i+1]))[0] - 1
        times = spike_times[indices]
        offset = np.floor(np.random.rand() * (stimulus_times[i+1] - start))
        offset_times = times + offset + start - i*(stimulus_times[i+1] - start)
        wrapped_times = np.array([b if b < stimulus_times[i+1] else b-(stimulus_times[i+1]-start) for b in offset_times])
        #print str(start)+' '+str(offset)+' '+str(offset_times)+' '+str(wrapped_times)

        shifted_times[indices] = wrapped_times
    indices = np.where(spike_times > stimulus_times[-1])[0]
    times = spike_times[indices]
    offset = np.random.rand() * (spike_times[-1] - stimulus_times[-1])
    offset_times = times + offset
    wrapped_times = np.array([b if b < spike_times[-1] else b-(spike_times[-1]-stimulus_times[-1]) for b in offset_times])
    shifted_times[indices] = wrapped_times
    return np.sort(shifted_times) - stimulus_times[0]


# def Rjitter_pair(spike_times, other_spike times):
#     spearmanr(spike_times,other_spike,nan_policy='omit')
    



from distutils.core import setup

setup(name='denmanlab',
      version='0.4.1',
      description='denmanlab code, mostly for neurophysiology',
      author='denman lab',
      author_email='daniel.denman@cuanschutz.edu',
      url='https://denmanlab.github.io',
      packages= ['dlab'],
      install_requires = ['scipy', 'matplotlib', 'numpy','pandas','scikit-learn'],
     )


from brainrender import Scene
from brainrender import settings
from brainrender.actors import Points
settings.SHOW_AXES = False
settings.WHOLE_SCREEN = False

import numpy as np
from matplotlib import cm
from matplotlib.colors import to_hex
import pandas as pd

#load data
path = '/Users/danieljdenman/Desktop/crossings.csv'
df = pd.read_csv(path).transpose()
df.columns = df.iloc[0]

#calculate channel coords relative to an insertion origin
#assumes 45º angle insertion (to the horizontal plane), and 120º angle between each (in the horizontal plane). 
channels_in = 300
x=700
x_ = channels_in*10 / np.sqrt(2)
a=(3000-x*np.sqrt(2))/np.sqrt(2)
stim_electrode = [np.zeros(channels_in),
                  np.zeros(channels_in),
                  np.linspace(500,0,channels_in),
                  ]
probeA         = [np.linspace(a/2. * np.sqrt(3),-(x/2.)*np.sqrt(3),channels_in),
                  np.linspace(x_,0,channels_in),
                  np.linspace(-a/2.,(x/2.),channels_in),
                  ]
probeB         = [np.linspace(-a/2. * np.sqrt(3),(x/2.)*np.sqrt(3),channels_in),
                  np.linspace(x_,0,channels_in),
                  np.linspace(-a/2.,(x/2.),channels_in),
                  ]
probeC         = [np.ones(channels_in),
                  np.linspace(x_,0,channels_in),
                  np.linspace(x_-x,-x,channels_in),
                  ]

#create brainrender scene and add relevant areas using Allen CCF names
scene = Scene()
scene.add_brain_region("VISp",alpha=0.1)
scene.add_brain_region("VISpl",alpha=0.1)
scene.add_brain_region("HIP",alpha=0.1)
scene.add_brain_region("SUB",alpha=0.1)
scene.add_brain_region("POST",alpha=0.1)
scene.add_brain_region("ENT",alpha=0.1)
scene.add_brain_region("SCs",alpha=0.1)
scene.add_brain_region("SCm",alpha=0.1)

#add the data from each channel as a colored point
insertion_origin = np.array([9000,400,3200]) #[AP,DV,ML]
data = ['jlh6a','jlh6b','jlh6c']
mP = cm.get_cmap('inferno',50)
for i,p in enumerate([probeA,probeB,probeC]):
    cs = [to_hex(mP(c/5000.)) for c in df[data[i]].values[1:].astype(float)]
    coords = np.array([c+insertion_origin for c in np.array(p).T.reshape(-1,3)])
    scene.add(Points(coords,colors=cs))

#render with specified view. to change the view, can interact with plot then press 'c'. copy the params printed to the console and paste them over the ones in custom_camera
scene.content
custom_camera =  {
     'pos': (-14214, 1547, 13938),
     'viewup': (0, -1, 0),
     'clippingRange': (10986, 50720),
     'focalPoint': (8645, 3895, -3625),
     'distance': 28923,
   }
scene.render(camera=custom_camera,zoom=2.)
